/*
 * mod_sentry.c - sentry module for Apache
 *
 *  by Information Network/Communication Laboratory, NAIST. 1996-2002.
 *      Masanao Izumo
 *      Ken-ichi Chinen
 *      Atushi Mino
 *
 * See. http://infonet.aist-nara.ac.jp/products/sentry/index.html
 *
 */
/*
 * NO WARRANTRY. OWN YOUR RISK.
 */

#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_main.h"
#include "http_protocol.h"
#include "util_script.h"

#include <stdio.h>
#include <sys/ipc.h>
#include <sys/shm.h>

#include <syslog.h>

#define MAX_RETRY_COUNT 10
#define DEFAULT_IMAGES 16
#define SENTRY_FAILURE_COUNT 8

module sentry_module;
static char sentry_root_dir[256] = "/var/tmp/sentry_save";
static char sentry_url_root[256] = "/bscr-imgs";
static int sentry_retr_delay = 2;
static char sentry_default_images[DEFAULT_IMAGES][256];
static int default_image_count;
static int sentry_max_refresh_images = 32;


/* SHM options */
#define SHM_SEGMENT_SIZE (1024*8) /* 8K */
#define SHM_SEGMENT_COUNT 257   /* 256 + 1(Status) */
#define SEGMENT_HDRSIZ 16
#define SEGMENT_DATSIZ (sentry_shm_segsiz - SEGMENT_HDRSIZ)
#if 0 
static key_t sentry_shm_key = -1;
static long sentry_shm_segsiz, sentry_shm_segnum;
static int sentry_shmid = -1;
#endif
volatile static char* sentry_shm_buffer = NULL;
static char* sentry_refresh_header = NULL;


typedef struct {
    key_t sentry_shm_key;
    long sentry_shm_segsiz, sentry_shm_segnum;
    unsigned char* sentry_shm_buffer;
    int sentry_shmid;
    int sentry_retr_delay;
    char* sentry_refresh_header;
    int recover_mode;
    int debug_mode;
} sentry_rec;


typedef unsigned int int32;     /* 32 bit integer */

/*
 * URL Naming schema
 *
 * /[^/]+/refresh/[0-9][0-9][0-9]/[0-9][0-9][0-9]/[0-9][0-9][0-9]/index.html
 *               /    Refresh    /   Interval    /    N Images   /
 *
 * or else
 *
 * /[^/]+/time/[0-9]+.html
 *             epochtime
 *
 * or else
 *
 * deliver the current image
 */

static const char*
set_sentry_refresh_header(cmd_parms *cmd, void *dummy, char *arg)
{
    int fd;
    struct stat s;

    if((fd = open(arg, O_RDONLY)) < 0)
        return strerror(errno);
    if(fstat(fd, &s) < 0)
    {
        close(fd);
        return strerror(errno);
    }
    if(sentry_refresh_header != NULL)
        free(sentry_refresh_header);
    if((sentry_refresh_header = (char *)malloc(s.st_size + 1)) == NULL)
    {
        char* err = strerror(errno);
        close(fd);
        return err;
    }

    read(fd, sentry_refresh_header, s.st_size);
    close(fd);
    sentry_refresh_header[s.st_size] = '\0';

    return NULL;
}

static const char*
set_sentry_retr_delay(cmd_parms *cmd, void *dummy, char *arg)
{
    sentry_retr_delay = atoi(arg);
    return NULL;
}

static const char*
set_sentry_root(cmd_parms *cmd, void *dummy, char *arg)
{
    if(!ap_is_directory (arg))
        return "SentryRoot must be a valid directory";
    strncpy(sentry_root_dir, arg, sizeof(sentry_root_dir) - 1);
    sentry_root_dir[sizeof(sentry_root_dir) - 1] = '\0';
    return NULL;
}

static const char*
set_sentry_url_root(cmd_parms *cmd, void *dummy, char *arg)
{
    strncpy(sentry_url_root, arg, sizeof(sentry_url_root) - 1);
    sentry_url_root[sizeof(sentry_url_root) - 1] = '\0';
    return NULL;
}

static const char*
add_sentry_default_image(cmd_parms *cmd, void *dummy, char *arg)
{
  int i;

  for(i = 0; i < default_image_count; i++)
    if(strcmp(arg, sentry_default_images[i]) == 0)
      return NULL;

    if(default_image_count == DEFAULT_IMAGES)
        return "AddSentryDefaultImage: Can't add any more images.";
    strncpy(sentry_default_images[default_image_count], arg,
            sizeof(sentry_default_images[0]));
    sentry_default_images[default_image_count][sizeof(sentry_default_images[0]) - 1] = '\0';
    default_image_count++;
    return NULL;
}

static const char*
set_sentry_recover(cmd_parms *cmd, void *dummy)
{
    sentry_rec *rec = (sentry_rec*) dummy;
#if 1
    syslog(LOG_ALERT, "set_sentry_recover: rec %p", rec);
#endif
    rec->recover_mode = 1;

    return NULL;
}

static const char*
set_sentry_debug(cmd_parms *cmd, void *dummy)
{
    sentry_rec *rec = (sentry_rec*) dummy;
#if 1
    syslog(LOG_ALERT, "set_sentry_debug: rec %p", rec);
#endif
    rec->debug_mode = 1;

    return NULL;
}

static const char*
set_sentry_shm(cmd_parms *cmd, void *dummy, char *v1, char *v2, char *v3)
{
    sentry_rec *rec;
#if 1
    syslog(LOG_ALERT, "set_sentry_shm: rec %p v '%s' '%s' '%s'",
        dummy, v1, v2, v3);
#endif
    rec = (sentry_rec*) dummy;

    if(rec->sentry_shm_buffer != NULL)
    {
#if 1
        if(rec->sentry_shm_key == atoi(v1) &&
          rec->sentry_shm_segsiz == atoi(v2) &&
          rec->sentry_shm_segnum == atoi(v3) - 1)
            return NULL;                /* No changes */
        shmdt((char *)rec->sentry_shm_buffer);
        rec->sentry_shm_buffer = NULL;
#else
        return NULL;
#endif
    }

    rec->sentry_shm_key     = atoi(v1);
    rec->sentry_shm_segsiz  = atoi(v2);
    rec->sentry_shm_segnum  = atoi(v3)-1;

    if(rec->sentry_shm_key <= 0 ||
      rec->sentry_shm_segsiz <= 0 ||
      rec->sentry_shm_segnum <= 0)
        return "Invalid arguments";

    if((rec->sentry_shmid =
        shmget(rec->sentry_shm_key,
            rec->sentry_shm_segsiz * (1 + rec->sentry_shm_segnum), 0666)) < 0)
    {
        if((rec->sentry_shmid =
            shmget(rec->sentry_shm_key,
                rec->sentry_shm_segsiz * (1 + rec->sentry_shm_segnum),
                    IPC_CREAT | 0666)) < 0)
        {
            rec->sentry_shm_buffer = NULL;
            return strerror(errno);
        }
    }

    if((rec->sentry_shm_buffer = (unsigned char*)
        shmat(rec->sentry_shmid, NULL, 0)) == (char *)-1)
    {
        rec->sentry_shm_buffer = NULL;
        return strerror(errno);
    }

    return NULL;
}

static const char*
set_sentry_max_refresh_images(cmd_parms *cmd, void *dummy, char* arg)
{
  sentry_max_refresh_images = atoi(arg);
  if(sentry_max_refresh_images <= 0)
    return "Invalid argument";
  return NULL;
}



static char*
sentry_time_path(const char* root, time_t t)
{
    static char path[BUFSIZ];
    struct tm *timeptr;

    timeptr = localtime(&t);
    sprintf(path, "%s/%02d%02d%02d/%02d/%02d/%02d-00.jpeg",
            root ? root : "",
            timeptr->tm_year,
            timeptr->tm_mon + 1,
            timeptr->tm_mday,
            timeptr->tm_hour,
            timeptr->tm_min,
            timeptr->tm_sec);
    return path;
}

static int sentry_not_found(request_rec *r)
{
    ap_log_reason("File does not exist",
        r->path_info ? ap_pstrcat(r->pool, r->filename, r->path_info, NULL)
               : r->filename, r);
    return NOT_FOUND;
}

static int sentry_refresh_html_request(request_rec *r, char *name, char* arg)
{
    time_t t;
    int h_refresh, h_interval, h_nimages;
    int i;
    sentry_rec *conf =
        (sentry_rec*)ap_get_module_config(r->per_dir_config, &sentry_module);


#if 1
    syslog(LOG_ALERT,
        "sentry_refresh_html_request: conf %p (key %d), name '%s', arg '%s'",
        conf, conf->sentry_shm_key, name, arg);
#endif

    r->content_type = "text/html";
    ap_send_http_header(r);

#if 0 
    ap_rputs("--- lady and gentalman ---\n", r);
    ap_rprintf(r, "arg='%s'\n", arg);
#endif

    /* For HEAD Request */
    if(r->header_only)
    {
        ap_kill_timeout(r);
        return OK;
    }

    t = time(NULL);

    /* parse the arg */
    h_refresh = atoi(arg);
    while(*arg && *arg != '/')
        arg++;
    if(!*arg)
        return sentry_not_found(r);
    while(*arg == '/')
        arg++;

    h_interval = atoi(arg);
    while(*arg && *arg != '/')
        arg++;
    if(!*arg)
        return sentry_not_found(r);
    while(*arg == '/')
        arg++;

    h_nimages = atoi(arg);
    while(*arg && *arg != '/')
        arg++;
    if(!*arg)
        return sentry_not_found(r);
    while(*arg == '/')
        arg++;

    if(h_refresh <= 0 || h_interval <= 0 || h_nimages <= 0)
        return sentry_not_found(r);

    if(h_nimages > sentry_max_refresh_images)
        return sentry_not_found(r);

    ap_soft_timeout("send Sentry HTML", r);

    if(conf->sentry_refresh_header)
      {
        ap_rprintf(r, conf->sentry_refresh_header, h_refresh);
      }
    else
      {
        ap_rputs("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2//EN\">\n", r);
        ap_rputs("<HTML>\n", r);
        ap_rputs(" <HEAD>\n", r);
        ap_rprintf(r, " <META HTTP-EQUIV=\"refresh\" CONTENT=\"%d\">\n",
                h_refresh);
        ap_rputs("  <TITLE>Refresh Page</TITLE>\n", r);
        ap_rputs(" </HEAD>\n", r);
        ap_rputs("<BODY BGCOLOR=\"#D2F6C6\">\n", r);
#if 0
        ap_rputs("<B>Live pictures</B><BR>\n", r);
#endif
      }
    ap_rputs("<TABLE BORDER=0>\n", r);

    for(i = 0; i < h_nimages; i++)
    {
#if 0
        ap_rprintf(r, "<TR><TD><IMG SRC=\"%s/time/%d.jpeg\">\n",
                conf->sentry_url_root, t);
#endif
        ap_rprintf(r, "<TR><TD><IMG SRC=\"%s/time/%d.jpeg\">\n",
                name, t);
        t -= h_interval;
    }

    ap_rputs("</TABLE>\n", r);
    ap_rputs("</BODY>\n", r);
    ap_rputs("</HTML>\n", r);

    ap_kill_timeout(r);

    return OK;
}

static int sentry_newrefresh_html_request(request_rec *r, char *name, char* arg)
{
    time_t t;
    int h_refresh, h_interval, h_nimages;
    int i, c;
    int dat_time, dat_size;
    volatile char* datp;
    char *tstr;
    int now;

    sentry_rec *conf =
        (sentry_rec*)ap_get_module_config(r->per_dir_config, &sentry_module);


#if 1
    syslog(LOG_ALERT,
        "sentry_newrefresh_html_request: conf %p (key %d), name '%s', arg '%s'",
        conf, conf->sentry_shm_key, name, arg);
#endif

    now = time(NULL);
    r->content_type = "text/html";
    ap_send_http_header(r);

#if 0 
    ap_rputs("--- lady and gentalman ---\n", r);
    ap_rprintf(r, "arg='%s'\n", arg);
#endif

    /* For HEAD Request */
    if(r->header_only)
    {
        ap_kill_timeout(r);
        return OK;
    }

    ap_rputs("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2//EN\">\n", r);
    ap_rputs("<HTML>\n", r);
    ap_rputs(" <HEAD>\n", r);
    ap_rprintf(r, " <META HTTP-EQUIV=\"refresh\" CONTENT=\"%d\">\n",
            h_refresh);
    ap_rputs("  <TITLE>Renew Page</TITLE>\n", r);
    ap_rputs(" </HEAD>\n", r);
    ap_rputs("<BODY BGCOLOR=\"#D2F6C6\">\n", r);

    if(conf->debug_mode) {
        ap_rprintf(r, "Time: %d, <B>Key %d</B><BR>\n",
            now, conf->sentry_shm_key);
    }
#if 0
    ap_rputs("<TABLE BORDER=0>\n", r);
#endif


    c = 0;
    for(i=0;i<conf->sentry_shm_segnum;i++) {
        datp = conf->sentry_shm_buffer + (1 + i) * conf->sentry_shm_segsiz;
        dat_time = (int)((int32 *)datp)[0];
        dat_size = (int)((int32 *)datp)[1];
        if(dat_size>0) {
            tstr = ctime((const time_t*)&dat_time);
            ap_rprintf(r,
    "%d-%d <IMG SRC=\"%s/time/%d.jpeg\" WIDTH=\"160\" HEIGHT=\"120\"> %d %d %s<BR>\n",
                c, i, name, dat_time, dat_time, now-dat_time, tstr);
#if 1
    syslog(LOG_ALERT, "img %3d: %3d %p %5d %d", c, i, datp, dat_size, dat_time);
#endif
            c++;

        }
    }


#if 0
    ap_rputs("</TABLE>\n", r);
#endif
    ap_rputs("</BODY>\n", r);
    ap_rputs("</HTML>\n", r);

    ap_kill_timeout(r);

    return OK;
}


static FILE* sentry_default_image(request_rec *r)
{
    int i;
    FILE* fp;

#ifdef __EMX__
    /* Need binary mode for OS/2 */
    fp = ap_pfopen(r->pool, r->filename, "rb");
#else
    fp = ap_pfopen(r->pool, r->filename, "r");
#endif

    if(fp != NULL)
      return fp;

    for(i = 0; i < default_image_count; i++)
    {
#ifdef __EMX__
        /* Need binary mode for OS/2 */
        fp = ap_pfopen(r->pool, sentry_default_images[i], "rb");
#else
        fp = ap_pfopen(r->pool, sentry_default_images[i], "r");
#endif

        if(fp != NULL)
            return fp;
    }
    return NULL;
}

static int sentry_maybe_down(request_rec *r)
{
    char* file;

    file = r->filename;
    if(file == NULL)
        file = r->uri;

    ap_log_reason("No sentry file",
               r->path_info ? ap_pstrcat(r->pool, r->filename, r->path_info, NULL)
               : file, r);
    return NOT_FOUND;
}

static int
send_bytes_length(const char* buf, request_rec *r, long length)
{
    while(length > 0 && !r->connection->aborted)
    {
        int len;
#if 1 
        printf("## length=%d\n", length);
#endif
        len = IOBUFSIZE;
        if(len > length)
            len = length;

        while(len && !r->connection->aborted)
        {
            int w;

            w = ap_bwrite(r->connection->client, buf, len);
            if(w > 0)
            {
                ap_reset_timeout(r); /* reset timeout after successful write */
                len -= w;
                buf += w;
                length -= w;
            }
            else if(w < 0)
            {
                if(r->connection->aborted)
                    break;
                else if(errno == EAGAIN)
                    continue;
                else
                {
                    ap_log_unixerr("send JPEG lost connection to",
                                ap_get_remote_host(r->connection,
                                                r->per_dir_config,
                                                REMOTE_NAME),
                                NULL, r->server);
                    ap_bsetflag(r->connection->client, B_EOUT, 1);
                    r->connection->aborted = 1;
                    break;
                }
            }
        }
    }

    if(r->connection->aborted)
        return EOF;
    if(r->sent_bodyct)
        ap_bgetopt (r->connection->client, BO_BYTECT, &r->bytes_sent);
    return 1;
}


int sentry_shm_send_image(request_rec *r, time_t t)
{
    int i;
    int alt_t;
    int now;
#if 0
    int32 dat_time, dat_size;
#endif
    int dat_time, dat_size;
    volatile char* datp;
    int retry_cnt;
    int most_fresh;
    sentry_rec *conf =
        (sentry_rec*)ap_get_module_config(r->per_dir_config, &sentry_module);

#if 1
    syslog(LOG_ALERT, "sentry_shm_send_image: conf %p (key %d) t %d",
        conf, conf->sentry_shm_key, t);
#endif

    retry_cnt = MAX_RETRY_COUNT;
    r->content_type = "image/jpeg";
    t -= conf->sentry_retr_delay;

    /* CASE 1:  */
    do {
        datp = conf->sentry_shm_buffer +
            (1 + t % conf->sentry_shm_segnum) * conf->sentry_shm_segsiz;
        dat_time = (int)((int32 *)datp)[0];
        dat_size = (int)((int32 *)datp)[1];
#if 0
        syslog(LOG_ALERT, "       %d %#x : %3d: %p %5d %d %6d\n",
            conf->sentry_shm_key, conf->sentry_shm_key,
            retry_cnt, datp, dat_size, dat_time, t - dat_time);
#endif

        if(dat_time == t)
            break;
/*
        t--;
*/
        retry_cnt--;
    } while(retry_cnt > 0);


    if(conf->recover_mode) {

    now = time(NULL);
    most_fresh = -1;

    for(i=0;i<conf->sentry_shm_segnum;i++) {
        datp = conf->sentry_shm_buffer + (1 + i) * conf->sentry_shm_segsiz;
        dat_time = (int)((int32*)datp)[0];
        dat_size = (int)((int32*)datp)[1];

        if(dat_time>most_fresh) {
            most_fresh = dat_time;
        }
#if 0
        if(dat_size) {
        syslog(LOG_ALERT, " fresh %d %#x : %3d: %p %5d %d %d (%d)\n",
            conf->sentry_shm_key, conf->sentry_shm_key,
            i, datp, dat_size, dat_time, most_fresh, dat_time>most_fresh );
        }
#endif
    }

    alt_t = t - (now - most_fresh);
#if 1
        syslog(LOG_ALERT, " guess alt_t %d ; now %d fresh %d t %d, gap %d",
            alt_t, now, most_fresh, t, now - most_fresh);
#endif

    for(i=0;i<conf->sentry_shm_segnum;i++) {
        datp = conf->sentry_shm_buffer + (1 + i) * conf->sentry_shm_segsiz;
        dat_time = (int)((int32 *)datp)[0];
        dat_size = (int)((int32 *)datp)[1];
#if 0
        if(dat_size) {
        syslog(LOG_ALERT, " guess %d %#x : %3d: %p %5d %d %6d\n",
            conf->sentry_shm_key, conf->sentry_shm_key,
            i, datp, dat_size, dat_time, alt_t - dat_time);
        }
#endif

        if(dat_time <= alt_t)
            break;
/*
        t--;
*/
    }
    if(dat_time > alt_t)
        return sentry_maybe_down(r);

    }
    else {
        
    if(dat_time != t)
        return sentry_maybe_down(r);
    }


    ap_send_http_header(r);

    if(dat_size > conf->sentry_shm_segsiz)
      {
#if 1 
        printf("## Overflow\n");
#endif
        dat_size = conf->sentry_shm_segsiz;
      }
#if 1
    syslog(LOG_ALERT, "sentry_shm_send_image: t %d, datp %p, dat_time %d, dat_size %d",
        t, datp, dat_time, dat_size);
    syslog(LOG_ALERT, "D: key %d (%#x), base %d %p, datp %p",
        conf->sentry_shm_key, conf->sentry_shm_key,
        conf->sentry_shm_buffer, conf->sentry_shm_buffer, datp);
#endif

    send_bytes_length((char *)datp + SEGMENT_HDRSIZ, r, dat_size);

    return OK;
}

int sentry_send_image(request_rec *r, time_t t_arg)
{
    char* path;
    int retry_cnt;
    FILE* fp;
    time_t t;
    static time_t failure_count = 0;
    sentry_rec *conf =
        (sentry_rec*)ap_get_module_config(r->per_dir_config, &sentry_module);

    if(t_arg == 0)
        t = time(NULL);
    else
        t = t_arg;

    if(conf->sentry_shm_buffer != NULL)
    {
        return sentry_shm_send_image(r, t);
    }

    retry_cnt = MAX_RETRY_COUNT;
    r->content_type = "image/jpeg";
    t -= sentry_retr_delay;

    do
    {
        if(retry_cnt == MAX_RETRY_COUNT - 1 &&
           failure_count > SENTRY_FAILURE_COUNT)
        {
            if((fp = sentry_default_image(r)) == NULL)
                return sentry_maybe_down(r);
            break;
        }
        path = sentry_time_path(sentry_root_dir, t);

#ifdef __EMX__
        /* Need binary mode for OS/2 */
        fp = ap_pfopen (r->pool, path, "rb");
#else
        fp = ap_pfopen (r->pool, path, "r");
#endif

        t--;
        retry_cnt--;

    } while(fp == NULL && retry_cnt > 0);

    if(fp == NULL)
    {
        if((fp = sentry_default_image(r)) == NULL)
        {
            failure_count++;
            return sentry_maybe_down(r);
        }
    }

    if(t_arg == 0)
    {
        failure_count = 0;
    }

    ap_send_http_header(r);

    if(!r->header_only)
        ap_send_fd(fp, r);
    ap_pfclose(r->pool, fp);
    return OK;
}



static int sentry_setfile(request_rec *r)
{
    if(sentry_shm_buffer != NULL &&
       strncmp(r->uri, sentry_url_root, strlen(sentry_url_root)) == 0)
        return OK;
    return DECLINED;
}


static void sentry_init(server_rec *r, pool *p)
{
    ap_add_version_component("sentry-red/1.0");
}

static void* sentry_dir(pool *p, char *dir)
{
    sentry_rec *cur;

    cur = (sentry_rec*)ap_palloc(p, sizeof(sentry_rec));
    cur->sentry_shm_key     = -1;
    cur->sentry_shm_segsiz  = -1;
    cur->sentry_shm_segnum  = -1;
    cur->sentry_shm_buffer  = NULL;
    cur->sentry_shmid       = -1;
    cur->sentry_refresh_header = NULL;
    cur->sentry_retr_delay  = 1;

    cur->recover_mode = 0;
    cur->debug_mode = 0;
#if 0
#endif

#if 1
    syslog(LOG_ALERT, "sentry_dir: '%s' as %p (key %d)",
        dir, cur, cur->sentry_shm_key);
#endif
    return cur;
}

#if 0
static void* sentry_dir_merge(pool *p, void *base_conf, void *sub_conf)
{
    sentry_rec *parent, *sub, *cur;
    
    parent = (sentry_rec*)base_conf;
    sub = (sentry_rec*)sub_conf;
    cur = (sentry_rec*)ap_palloc(p, sizeof(sentry_rec));

#if 1
    syslog(LOG_ALERT, "sentry_dir_merge: parent %p (key %d), sub %p (key %d) => cur %p (key %d)",
        parent, parent->sentry_shm_key,
        sub, sub->sentry_shm_key,
        cur, cur->sentry_shm_key);
#endif

    return cur;
}
#endif


static int sentry_handler(request_rec *r);

static command_rec sentry_cmds[] = {
{"SentryRoot", set_sentry_root, NULL, ACCESS_CONF, TAKE1,
     "Set sentry spool root directory" },
{"SentryURLRoot", set_sentry_url_root, NULL, ACCESS_CONF, TAKE1,
     "Set sentry URL root directory" },
{"SentryRetrDelay", set_sentry_retr_delay, NULL, ACCESS_CONF, TAKE1,
     "Set sentry retrival delay" },
{"AddSentryDefaultImage", add_sentry_default_image, NULL, ACCESS_CONF, TAKE1,
     "Add new sentry default image"},
{"SentrySHM", set_sentry_shm, NULL, ACCESS_CONF, TAKE3,
     "SentrySHM KEY FRAME-SIZE FRAME-NUM"},
{"SentryMAXRefreshImages", set_sentry_max_refresh_images,
   NULL, ACCESS_CONF, TAKE1,
   "Set Sentry Max Refresh Image No." },
{"SentryRefreshHeader", set_sentry_refresh_header, NULL, ACCESS_CONF, TAKE1,
   "Set Sentry Refresh Redirect"},
{"SentryRecover", set_sentry_recover, NULL, ACCESS_CONF, NO_ARGS,
     "Enable sentry recover" },
{"SentryDebug", set_sentry_debug, NULL, ACCESS_CONF, NO_ARGS,
     "Enable sentry debug" },
{NULL}
};

static handler_rec sentry_handlers[] = {
{ "sentry", sentry_handler },
{ NULL }
};

#if 0
    void (*init) (server_rec *, pool *);
    void *(*create_dir_config) (pool *p, char *dir);
    void *(*merge_dir_config) (pool *p, void *base_conf, void *new_conf);
    void *(*create_server_config) (pool *p, server_rec *s);
    void *(*merge_server_config) (pool *p, void *base_conf, void *new_conf);
#endif

module sentry_module = {
    STANDARD_MODULE_STUFF,
    sentry_init,        /* initializer */
    sentry_dir,         /* per-directory config creater */
#if 0
    sentry_dir_merge,   /* dir config merger - default is to override */
#endif
    NULL,               /* dir config merger - default is to override */
    NULL,               /* server config creator */
    NULL,               /* server config merger */
    sentry_cmds,        /* command table */
    sentry_handlers,    /* [6] list of handlers */
    sentry_setfile,     /* [1] filename-to-URI translation */
    NULL,               /* [4] check/validate HTTP user_id */
    NULL,               /* [5] check HTTP user_id is valid *here* */
    NULL,               /* [3] check access by host address, etc. */
    NULL,               /* [6] MIME type checker/setter */
    NULL,               /* [7] fixups */
    NULL,               /* [9] logger */
    NULL                /* [2] header parser */
};

static int sentry_handler(request_rec *r)
{
    char *uri_check;
    char sentry_id[BUFSIZ];
    char *q;
    int   c;
    sentry_rec *conf =
        (sentry_rec*)ap_get_module_config(r->per_dir_config, &sentry_module);
    
#if 1
    printf("## sentry_handler\n");
#endif

#if 1
    syslog(LOG_ALERT, "URI '%s', config %p (key %d)",
        r->uri, conf, conf->sentry_shm_key);
#endif

    if(r->method_number != M_GET)
        return METHOD_NOT_ALLOWED;

    uri_check = r->uri;

    /*
     * skipping / * /
     */

    if(*uri_check != '/')
        return sentry_not_found(r);
    while(*uri_check == '/')
        uri_check++;

    c = 0;
    q = sentry_id;
    *q++ = '/';
    while(*uri_check && *uri_check != '/') {
        if(c<BUFSIZ-1) {
            *q++ = *uri_check;
        }
        uri_check++;
    }
    *q = '\0';

    if(*uri_check != '/')
        return sentry_not_found(r);
    while(*uri_check == '/')
        uri_check++;

#if 1
    syslog(LOG_ALERT, "URI check 0 '%s' + '%s'\n",
        sentry_id, uri_check);
#endif

    if(strncmp(uri_check, "refresh/", 8) == 0) {
#if 0
        syslog(LOG_ALERT, "refresh\n");
#endif
        return sentry_refresh_html_request(r, sentry_id, uri_check + 8);
    }

    if(strncmp(uri_check, "renew/", 6) == 0) {
#if 1
        syslog(LOG_ALERT, "renew\n");
#endif
        return sentry_newrefresh_html_request(r, sentry_id, uri_check + 6);
    }

    if(strncmp(uri_check, "time/", 5) == 0) {
#if 0
        syslog(LOG_ALERT, "time\n");
#endif
        return sentry_send_image(r, atoi(uri_check + 5));
    }

    return sentry_send_image(r, 0);
}

